/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <ortp/ortp.h>
#include "recv.cmdline.h"

#define CHUNK_SIZE 65536

extern gint msg_to_buf(mblk_t *, char *, gint);


static RtpSession *session;
static struct gengetopt_args_info conf;
/* header_status < 0: need header, == 0: have header, > 0: don't need header */
static int header_len, header_status = -1;
enum {END_OF_FILE, BEGIN_OF_FILE, HEADER};
static char *header_buf;

void para_log(int, char*, ...) __attribute__((format (printf, 2, 3)));
void para_log(int ll, char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static int write_buffer(mblk_t *mp)
{
	char buf[CHUNK_SIZE];
	static int len;
	int ret = -1;

	if (!header_status && conf.header_given) {
		header_status = 1;
		PARA_INFO_LOG("writing header (%d bytes)\n", header_len);
		ret = write(STDOUT_FILENO, header_buf, header_len);
		if (ret != header_len) {
			ret = -1;
			goto out;
		}
		/*
                 * start playback in the middle of the file. This can't
                 * work reliably right now as we do not have any clue
                 * how much data we should buffer before sending. It is
                 * however not a serious problem as it usually affects
                 * only the very first file that is streamed. Just sleep
                 * a bit and hope that this prevents buffer underuns in
                 * the player.
		 */
		sleep(1);
	}
	if (!mp)
		goto write; /* eof but might still have some data in buffer */
	if (len >= CHUNK_SIZE)
		goto out;
	PARA_DEBUG_LOG("msg2buf: loaded: %d\n", len);
	ret = msg_to_buf(mp, buf + len, CHUNK_SIZE - len);
	if (ret <= 0)
		goto out;
	len += ret;
write:
	ret = -1;
	if (len > 0) {
		ret = write(STDOUT_FILENO, buf, len);
		header_status = 1;
		PARA_DEBUG_LOG("wrote %d/%d\n", ret, len);
		if (ret != len)
			return -1;
		len -= ret;
	}
	ret = 1;
out:
	if (ret <= 0)
		len = 0;
	return ret;
}

static int handle_other_package(mblk_t *mp)
{
	char buf[CHUNK_SIZE];
	int ret = -1, code, len = msg_to_buf(mp, buf, CHUNK_SIZE);

	if (len <= 0) {
		PARA_ERROR_LOG("msg_to_buf failed: %i\n", len);
		return -1;
	}
	code = buf[0];
	PARA_INFO_LOG("rcvd ctrl pkg #%i (%d bytes)\n", code, len);
	switch (code) {
	case END_OF_FILE:
		ret = 0;
		write_buffer(NULL);
		goto out;
		break;
	case BEGIN_OF_FILE:
		header_status = 1;
		break;
	case HEADER:
		if (!conf.header_given || header_status >= 0)
			break;
		header_status = 0;
		header_len = len - 1;
		header_buf = para_malloc((header_len));
		memmove(header_buf, buf + 1, header_len);
		PARA_NOTICE_LOG("have header: %d bytes\n", header_len);
		if (ret == len - 1)
			break;
		ret = -1;
		break;
	default:
		PARA_WARNING_LOG("%s", "rcvd unknown control pkg\n");
		goto out;
	}
	ret = 1;
out:
	freemsg(mp);
	return ret;
}

int main(int argc, char *argv[])
{
	int ret;
	unsigned time = 10;
	guint32 ts = 0;
	mblk_t *mp;

	cmdline_parser(argc, argv, &conf);
	ortp_init();
	ortp_scheduler_init();
	ortp_set_debug_file("oRTP", NULL);
	session = rtp_session_new(RTP_SESSION_RECVONLY);
	rtp_session_set_scheduling_mode(session, 1);
	rtp_session_set_blocking_mode(session, 1);
	PARA_NOTICE_LOG("receiving from %s:%d\n", conf.hostname_arg,
		conf.port_arg);
	rtp_session_set_local_addr(session, conf.hostname_arg, conf.port_arg);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
recv:
	mp = rtp_session_recvm_with_ts(session, ts);
	if (!mp) {
//		PARA_DEBUG_LOG("did not receive buffer, time = %i\n", time);
		if (time < 1000)
			time += 10;
		ts += time;
		goto recv;
	}
	if (rtp_session_get_payload_type(session) == PAYLOAD_OTHER) {
		ret = handle_other_package(mp);
		if (ret <= 0)
			goto out;
		goto recv;
	}
	if (header_status < 0 && conf.header_given) {
		PARA_DEBUG_LOG("%s", "have data, but no header\n");
		goto recv;
	}
	PARA_DEBUG_LOG("have data pkg, time = %u\n", time);
	ret = write_buffer(mp);
	freemsg(mp);
	if (ret <= 0)
		goto out;
	time = 10;
	goto recv;
out:
	PARA_NOTICE_LOG("exit code: %d\n", ret);
	rtp_session_destroy(session);
	ortp_exit();
	ortp_global_stats_display();
	return ret;
}
